"""Libchickadee

Yet another GeoIP resolution tool.
"""

__author__ = 'Chapin Bryce'
__date__ = 20200107
__license__ = 'GPLv3 Copyright 2019 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''



"""Chickadee script tests."""
import unittest
import os

from libchickadee.chickadee import Chickadee

__author__ = 'Chapin Bryce'
__date__ = 20200114
__license__ = 'GPLv3 Copyright 2019 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class ChickadeeStringTestCase(unittest.TestCase):
    """Chickadee script tests."""
    def setUp(self):
        """Test config"""
        self.test_data_ips = [
            '10.0.1.2', '8.8.8.8', '2001:4860:4860::8888'
        ]
        self.expected_result = [
            {'query': '10.0.1.2', 'count': 1},

            {'as': 'AS15169 Google LLC', 'country': 'United States',
             'org': 'Level 3', 'proxy': False, 'query': '8.8.8.8', 'count': 1},

            {'as': 'AS15169 Google LLC', 'country': 'United States',
             'org': 'Google LLC', 'proxy': False, 'count': 1,
             'query': '2001:4860:4860::8888'}
        ]

        self.fields = ['query', 'count', 'as', 'country', 'org', 'proxy']

    def test_no_resolve(self):
        results = [
            {'query': '10.0.1.2', 'count': 1, 'message': 'No resolve'},
            {'query': '8.8.8.8', 'count': 1, 'message': 'No resolve'},
            {'query': '2001:4860:4860::8888', 'count': 1,
             'message': 'No resolve'}
        ]
        for count, ip in enumerate(self.test_data_ips):
            chickadee = Chickadee()
            chickadee.fields = self.fields
            chickadee.resolve_ips = False
            data = chickadee.run(ip)
            res = [x for x in data]
            self.assertEqual(res, [results[count]])

    def test_chickadee_single(self):
        """Query Method Test"""
        for count, ip in enumerate(self.test_data_ips):
            chickadee = Chickadee()
            chickadee.fields = self.fields
            data = chickadee.run(ip)
            res = [x for x in data]
            self.assertEqual(res, [self.expected_result[count]])

    def test_chickadee_csv_str(self):
        """Batch Query Method Test"""
        chickadee = Chickadee()
        chickadee.fields = self.fields
        data = chickadee.run(','.join(self.test_data_ips))
        res = [x for x in data]
        self.assertCountEqual(res, self.expected_result)


class ChickadeeFileTestCase(unittest.TestCase):
    """Chickadee script tests."""
    def setUp(self):
        """Test setup."""
        self.test_data_dir = os.path.join(
            os.path.dirname(__file__), 'test_data')
        self.txt_data_results = [
            {"as": "AS15169 Google LLC",
             "country": "United States", "org": "Google LLC",
             "proxy": False, "query": "2001:4860:4860::8844",
             "count": 1},

            {"as": "AS15169 Google LLC",
             "country": "United States", "org": "Google LLC",
             "proxy": False, "query": "2001:4860:4860::8844",
             "count": 1},

            {"query": "10.0.1.2", "count": 1},

            {"as": "AS15169 Google LLC",
             "country": "United States", "org": "Level 3",
             "proxy": False, "query": "8.8.8.8", "count": 1},

            {"as": "AS15169 Google LLC",
             "country": "United States", "org": "Google LLC",
             "proxy": False, "query": "2001:4860:4860::8888",
             "count": 1},

            {"as": "AS3215 Orange S.A.",
             "country": "France", "org": "", "proxy": True,
             "query": "2.2.2.2", "count": 1},

            {"as": "AS15169 Google LLC",
             "country": "United States", "org": "Google LLC",
             "proxy": False, "query": "2001:4860:4860::8888", "count": 1},

            {"as": "AS3356 Level 3 Communications, Inc.",
             "country": "United States", "org": "Informs",
             "proxy": True, "query": "4.4.4.4", "count": 1},

            {"as": "AS13335 Cloudflare, Inc.",
             "country": "Australia", "org": "", "proxy": False,
             "query": "1.1.1.1", "count": 2}
        ]

        self.fields = ['query', 'count', 'as', 'country', 'org', 'proxy']

        self.xlsx_data_results = [
            {"as": "AS15169 Google LLC",
             "country": "United States", "org": "Google LLC",
             "proxy": False, "query": "2001:4860:4860::8844",
             "count": 2},

            {"query": "10.0.1.2", "count": 1},

            {"as": "AS15169 Google LLC",
             "country": "United States", "org": "Level 3",
             "proxy": False, "query": "8.8.8.8", "count": 1},

            {"as": "AS15169 Google LLC",
             "country": "United States", "org": "Google LLC",
             "proxy": False, "query": "2001:4860:4860::8888",
             "count": 1},

            {"as": "AS3215 Orange S.A.",
             "country": "France", "org": "", "proxy": True,
             "query": "2.2.2.2", "count": 1},

            {"as": "AS3356 Level 3 Communications, Inc.",
             "country": "United States", "org": "Informs",
             "proxy": True, "query": "4.4.4.4", "count": 1},

            {"as": "AS13335 Cloudflare, Inc.",
             "country": "Australia", "org": "", "proxy": False,
             "query": "1.1.1.1", "count": 2}
        ]

    def test_ipapi_resolve_query_txt_file(self):
        """Batch Query Method Test"""
        chickadee = Chickadee()
        chickadee.fields = self.fields
        data = chickadee.run(os.path.join(self.test_data_dir, 'txt_ips.txt'))
        res = [x for x in data]
        batch_result = []  # No reverse field
        for item in self.txt_data_results:
            if 'reverse' in item:
                item.pop('reverse')
            batch_result.append(item)
        self.assertCountEqual(res, batch_result)

    def test_ipapi_resolve_query_gz_file(self):
        """Batch Query Method Test"""
        chickadee = Chickadee()
        chickadee.fields = self.fields
        data = chickadee.run(os.path.join(self.test_data_dir,
                                          'txt_ips.txt.gz'))
        res = [x for x in data]
        batch_result = []  # No reverse field
        for item in self.txt_data_results:
            if 'reverse' in item:
                item.pop('reverse')
            batch_result.append(item)
        self.assertCountEqual(res, batch_result)

    def test_ipapi_resolve_query_xlsx_file(self):
        """Batch Query Method Test"""
        chickadee = Chickadee()
        chickadee.fields = self.fields
        data = chickadee.run(os.path.join(self.test_data_dir, 'test_ips.xlsx'))
        res = [x for x in data]
        batch_result = []  # No reverse field
        for item in self.xlsx_data_results:
            if 'reverse' in item:
                item.pop('reverse')
            batch_result.append(item)
        self.assertCountEqual(res, batch_result)

    def test_ipapi_resolve_query_folder(self):
        """Batch Query Method Test"""
        expected = [
            {"country": "Australia", "org": "",
             "as": "AS13335 Cloudflare, Inc.",
             "proxy": False, "query": "1.1.1.1", "count": 6},

            {"query": "10.0.1.2", "count": 3},

            {"country": "United States", "org": "Level 3",
             "as": "AS15169 Google LLC",
             "proxy": False, "query": "8.8.8.8", "count": 3},

            {"country": "United States", "org": "Google LLC",
             "as": "AS15169 Google LLC",
             "proxy": False, "query": "2001:4860:4860::8888", "count": 3},

            {"country": "United States", "org": "Informs",
             "as": "AS3356 Level 3 Communications, Inc.", "proxy": True,
             "query": "4.4.4.4",
             "count": 3},

            {"country": "United States", "org": "Google LLC",
             "as": "AS15169 Google LLC",
             "proxy": False, "query": "2001:4860:4860::8844", "count": 4},

            {"country": "United States", "org": "Google LLC",
             "as": "AS15169 Google LLC",
             "proxy": False, "query": "2001:4860:4860::8888", "count": 3},

            {"country": "France", "org": "", "as": "AS3215 Orange S.A.",
             "proxy": True, "query": "2.2.2.2", "count": 3},

            {"country": "United States", "org": "Google LLC",
             "as": "AS15169 Google LLC",
             "proxy": False, "query": "2001:4860:4860::8844", "count": 4}
        ]

        chickadee = Chickadee()
        chickadee.fields = self.fields
        data = chickadee.run(self.test_data_dir)
        res = [x for x in data]
        self.assertCountEqual(res, expected)


if __name__ == '__main__':
    unittest.main()



"""IP-API Backend Tests."""
import unittest

from libchickadee.backends.ipapi import Resolver

__author__ = 'Chapin Bryce'
__date__ = 20200114
__license__ = 'GPLv3 Copyright 2019 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class IPAPITestCase(unittest.TestCase):
    """IP-API Backend Tests."""
    def setUp(self):
        """Test config"""
        self.test_data_ips = [
            '10.0.1.2', '8.8.8.8', '2001:4860:4860::8888'
        ]
        self.expected_result = [
            {'query': '10.0.1.2'},

            {'as': 'AS15169 Google LLC', 'country': 'United States',
             'org': 'Level 3', 'proxy': False, 'query': '8.8.8.8'},

            {'as': 'AS15169 Google LLC', 'country': 'United States',
             'org': 'Google LLC', 'proxy': False,
             'query': '2001:4860:4860::8888'}
        ]
        self.resolver = Resolver(fields=['query', 'count', 'as',
                                         'country', 'org', 'proxy'])

    def test_ipapi_resolve_query_single(self):
        """Query Method Test"""
        for count, ip in enumerate(self.test_data_ips):
            data = self.resolver.query(ip)
            self.assertEqual(data, self.expected_result[count])

    def test_ipapi_resolve_query_batch(self):
        """Batch Query Method Test"""
        data = self.resolver.query(self.test_data_ips)
        res = [x for x in data]
        batch_result = []  # No reverse field
        for item in self.expected_result:
            if 'reverse' in item:
                item.pop('reverse')
            batch_result.append(item)
        self.assertCountEqual(res, batch_result)

    def test_ipapi_resolve_single(self):
        """Single Query Method Test"""
        for count, ip in enumerate(self.test_data_ips):
            self.resolver.data = ip
            data = self.resolver.single()
            self.assertEqual(data, self.expected_result[count])

    def test_ipapi_resolve_batch(self):
        """Batch Query Method Test"""
        self.resolver.data = self.test_data_ips
        data = self.resolver.batch()
        res = [x for x in data]
        self.assertCountEqual(res, self.expected_result)

    def test_ipapi_resolve_single_field(self):
        """Single Query Method Test"""
        for count, ip in enumerate(self.test_data_ips):
            self.resolver.data = ip
            self.resolver.fields = ['query', 'country', 'as']
            data = self.resolver.single()

            expected = {}
            for field in ['query', 'country', 'as']:
                if field not in data:
                    continue
                expected[field] = self.expected_result[count].get(field, None)
            self.assertEqual(data, expected)


'''
import os
from libchickadee.backends.ipapi import ProResolver
# Disabled - Unit test only run locally due to use of API key
class IPAPIProTestCase(unittest.TestCase):
    """IP-API Backend Tests."""
    def setUp(self):
        """Test config"""
        self.test_data_ips = [
            '10.0.1.2', '8.8.8.8', '1.1.1.1', '2.2.2.2', '2001:4860:4860::8888'
        ]
        self.expected_result = [
            {'message': 'private range', 'query': '10.0.1.2'},

            {'as': 'AS15169 Google LLC', 'city': 'Ashburn',
             'country': 'United States', 'district': '', 'lat': 39.0438,
             'lon': -77.4874, 'mobile': False, 'org': 'Google Inc.',
             'proxy': False, 'query': '8.8.8.8', 'reverse': 'dns.google',
             'regionName': 'Virginia', 'zip': '20149'},

            {'as': 'AS13335 Cloudflare, Inc.', 'city': 'Sydney',
             'country': 'Australia', 'district': '', 'lat': -33.8688,
             'lon': 151.209, 'mobile': False, 'org': '', 'proxy': False,
             'query': '1.1.1.1', 'regionName': 'New South Wales',
             'zip': '1001', 'reverse': 'one.one.one.one'},

            {'as': 'AS3215 Orange S.A.', 'city': 'Aubervilliers',
             'country': 'France', 'district': '', 'lat': 48.9123,
             'lon': 2.38405, 'mobile': False, 'org': '', 'proxy': True,
             'query': '2.2.2.2', 'regionName': 'Île-de-France',
             'zip': '93300', 'reverse': ''},

            {'as': 'AS15169 Google LLC', 'city': 'Newark',
             'country': 'United States', 'district': '', 'lat': 40.7357,
             'lon': -74.1724, 'mobile': False, 'org': 'Google LLC',
             'proxy': False, 'query': '2001:4860:4860::8888',
             'regionName': 'New Jersey', 'zip': '07175',
             'reverse': 'dns.google',}
        ]
        self.resolver = ProResolver(os.environ.get('CHICKADEE_API_KEY', ''))

    def test_ipapi_resolve_query_single(self):
        """Query Method Test"""
        for count, ip in enumerate(self.test_data_ips):
            data = self.resolver.query(ip)
            res = [x for x in data]
            self.assertEqual(res, [self.expected_result[count]])

    def test_ipapi_resolve_query_batch(self):
        """Batch Query Method Test"""
        data = self.resolver.query(self.test_data_ips)
        res = [x for x in data]
        batch_result = [] # No reverse field
        for item in self.expected_result:
            if 'reverse' in item:
                item.pop('reverse')
            batch_result.append(item)
        self.assertCountEqual(res, batch_result)

    def test_ipapi_resolve_single(self):
        """Single Query Method Test"""
        for count, ip in enumerate(self.test_data_ips):
            self.resolver.data = ip
            data = self.resolver.single()
            res = [x for x in data]
            self.assertEqual(res, [self.expected_result[count]])

    def test_ipapi_resolve_batch(self):
        """Batch Query Method Test"""
        self.resolver.data = self.test_data_ips
        data = self.resolver.batch()
        res = [x for x in data]
        batch_result = [] # No reverse field
        for item in self.expected_result:
            if 'reverse' in item:
                item.pop('reverse')
            batch_result.append(item)
        self.assertCountEqual(res, batch_result)
# '''


if __name__ == '__main__':
    unittest.main()



"""Base class for all backends."""
import json
import csv

__author__ = 'Chapin Bryce'
__date__ = 20200107
__license__ = 'GPLv3 Copyright 2019 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class ResolverBase(object):
    """Generic base class for use by other backends."""
    def __init__(self, fields=list(), lang='en'):
        self.uri = None
        self.lang = lang
        self.supported_langs = []
        self.fields = fields  # Ordered list of fields to gather
        self.pbar = False  # Enable progress bars
        self.data = None

    def single(self):
        """Base method to handle single queries"""
        raise NotImplementedError()

    def batch(self):
        """Base method to handle batch queries"""
        raise NotImplementedError()

    def query(self, data):
        """Generic query handler to decide to use single or batch
            method for querying.

        Args:
            data (list, tuple,set or str): One or more IPs to resolve

        Return:
            (yield) request data iterator
        """
        self.data = data
        if isinstance(data, (list, tuple, set)):
            return self.batch()
        if isinstance(data, str):
            return self.single()

        raise NotImplementedError()

    @staticmethod
    def write_csv(outfile, data, headers=None):
        """Writes a list of dictionaries to a CSV file.

        Arguments:
            outfile (str): Path to output file
            data (list): List of dictionaries to write to file
            headers (list): Header row to use. If empty, will use the
                first dictionary in the `data` list.
        """

        if not headers:
            # Use the first line of data
            headers = [str(x) for x in data[0].keys()]

        if isinstance(outfile, str):
            open_file = open(outfile, 'w', newline="")
        else:
            open_file = outfile

        # Write only provided headers, ignore others
        csvfile = csv.DictWriter(open_file, headers,
                                 extrasaction='ignore')
        csvfile.writeheader()
        csvfile.writerows(data)

    @staticmethod
    def write_json(outfile, data, headers=None, lines=False):
        """Writes output in JSON format

        Args:
            outfile (str or fileobj): Path to or already open file
            data (list): List of dictionaries containing resolved data
            lines (bool): Whether to export 1 dictionary object per line or
                a whole json object.
        """
        if isinstance(outfile, str):
            open_file = open(outfile, 'w', newline="")
        else:
            open_file = outfile

        if headers:
            # Only include fields in headers
            # Include headers with no value if not present in original
            selected_data = []
            for x in data:
                d = {}
                for k, v in x.items():
                    if k in headers:
                        d[k] = v
                for h in headers:
                    if h not in d:
                        d[h] = None
                selected_data.append(d)
            data = selected_data

        if lines:
            for entry in data:
                open_file.write(json.dumps(entry)+"\n")
        else:
            open_file.write(json.dumps(data))



"""Parse IP addresses from plain text files
and feed to the Chickadee GeoIP API

Plain text files include logs, csvs, json, and other formats where ascii
strings contain IPv4 or IPv6 addresses.
"""

import binascii
import os
from gzip import GzipFile

from libchickadee.parsers import IPv4Pattern, IPv6Pattern, strip_ipv6

__author__ = 'Chapin Bryce'
__date__ = 20200114
__license__ = 'GPLv3 Copyright 2019 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class PlainTextParser(object):
    """Class to extract IP addresses from plain text
        and gzipped plain text files."""
    def __init__(self):
        self.ips = dict()

    @staticmethod
    def is_gz_file(filepath):
        """Validate whether the input is GZipped."""
        with open(filepath, 'rb') as test_f:
            return binascii.hexlify(test_f.read(2)) == b'1f8b'

    def parse_file(self, file_entry, is_stream=False):
        """Parse contents of the file."""
        if not is_stream:
            if self.is_gz_file(file_entry):
                file_data = GzipFile(filename=file_entry)
            else:
                file_data = open(file_entry, 'rb')
        else:
            if binascii.hexlify(file_entry.buffer.read(2)) == b'1f8b':
                file_data = GzipFile(fileobj=file_entry)
            else:
                file_data = file_entry.buffer

        for raw_line in file_data:
            line = raw_line.decode()
            for ipv4 in IPv4Pattern.findall(line):
                if ipv4 not in self.ips:
                    self.ips[ipv4] = 0
                self.ips[ipv4] += 1
            for ipv6 in IPv6Pattern.findall(line):
                if strip_ipv6(ipv6) not in self.ips:
                    self.ips[strip_ipv6(ipv6)] = 0
                self.ips[strip_ipv6(ipv6)] += 1

        if 'closed' in dir(file_data) and not file_data.closed:
            file_data.close()


if __name__ == "__main__":  # pragma: no cover
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('path', help="File or folder to parse")
    args = parser.parse_args()

    ptparser = PlainTextParser()
    if os.path.isdir(args.path):
        for root, _, files in os.walk(args.path):
            for fentry in files:
                ptparser.parse_file(os.path.join(root, fentry))
    else:
        ptparser.parse_file(args.path)
    print("{} unique IPs discovered".format(len(ptparser.ips)))



"""Installer for chickadee"""
import setuptools

with open('README.md') as fh:
    long_description = fh.read()

setuptools.setup(
    name='chickadee',
    version=20200114,
    description='Yet another GeoIP resolution tool.',
    author='Chapin Bryce',
    author_email='python@chapinb.com',
    url='https://github.com/chapinb/chickadee',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'chickadee = libchickadee.chickadee:entry'
        ]
    },
    install_requires=[
        "requests >= 2.22.0",
        "openpyxl >= 2.6.3",
        "tqdm >= 4.36.1"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ]
)



"""Backend leveraging the ip-api.com JSON API."""
import logging
import time
from datetime import datetime, timedelta

import requests
from tqdm import trange

from . import ResolverBase

logger = logging.getLogger('libchickadee.chickadee')

__author__ = 'Chapin Bryce'
__date__ = 20200114
__license__ = 'GPLv3 Copyright 2019 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''

FIELDS = [  # Ordered list of fields to gather
    'query',
    'as', 'org', 'isp'
    'continent', 'country', 'regionName', 'city',
    'district', 'zip',
    'mobile', 'proxy', 'reverse',
    'lat', 'lon', 'timezone'
    'status', 'message'
]


class Resolver(ResolverBase):
    """Class to handle ip-api.com API queries for IP addresses."""
    def __init__(self, fields=FIELDS, lang='en'):
        self.supported_langs = [
            'en', 'de', 'es', 'pt-BR', 'fr', 'ja', 'zh-CN', 'ru'
        ]
        if lang not in self.supported_langs:
            lang = 'en'
        ResolverBase.__init__(self, fields=fields, lang='en')

        self.uri = 'http://ip-api.com/'
        self.api_key = None
        self.enable_sleep = True
        self.wait_time = datetime.now()

    def rate_limit(self, headers):
        """Method to check rate limiting."""
        if int(headers.get('X-Rl', '0')) < 1:
            self.wait_time = datetime.now() + \
                timedelta(seconds=int(headers.get('X-Ttl', '0'))+1)

    def sleeper(self):
        """Method to sleep operations for rate limiting."""
        while True:
            now = datetime.now()
            wait_time = self.wait_time - now
            if wait_time.total_seconds() < 0:
                self.wait_time = datetime.now()
                return
            wt_sec = wait_time.total_seconds()+1  # add a buffer
            logger.info(
                'Sleeping for {} seconds due to rate limiting.'.format(wt_sec))
            time.sleep(wt_sec)

    def batch(self):
        """Handle batch query operations."""
        records = []
        for ip in self.data:
            records.append({'query': ip})

        resolved_recs = []
        if self.pbar:
            orig_recs = trange(0, len(records), 100,
                               desc="Resolving IPs", unit_scale=True)
        else:
            orig_recs = range(0, len(records), 100)

        for x in orig_recs:
            params = {
                'fields': ','.join(self.fields),
                'lang': self.lang,
            }
            if self.api_key:
                params['key'] = self.api_key

            if self.enable_sleep:
                self.sleeper()

            rdata = requests.post(
                self.uri+"batch",
                json=records[x:x+100],
                params=params
            )

            if rdata.status_code == 200:
                self.rate_limit(rdata.headers)
                result_list = [x for x in rdata.json()]
                resolved_recs += result_list
            elif rdata.status_code == 429:
                self.rate_limit(rdata.headers)
                self.sleeper()
                return self.batch()
            else:
                msg = "Unknown error encountered: {}".format(rdata.status_code)
                logger.error(msg)
                result_list = []
                for result in records[x:x+100]:
                    result_list.append({'query': result,
                                        'status': 'failed',
                                        'message': msg})
                resolved_recs += result_list
        return resolved_recs

    def single(self):
        """Handle single item query operations."""
        params = {
            'fields': ','.join(self.fields),
            'lang': self.lang,
        }
        if self.api_key:
            params['key'] = self.api_key

        if self.enable_sleep:
            self.sleeper()

        rdata = requests.get(
            self.uri+"json/"+self.data,
            params=params
        )
        if rdata.status_code == 200:
            self.rate_limit(rdata.headers)
            return rdata.json()
        elif rdata.status_code == 429:
            self.rate_limit(rdata.headers)
            self.sleeper()
            return self.single()
        else:
            msg = "Unknown error encountered: {}".format(rdata.status_code)
            logger.error(msg)
            return [{'query': self.data, 'status': 'failed', 'message': msg}]


class ProResolver(Resolver):
    """GeoIP resolver using the ip-api.com paid subscription."""
    def __init__(self, api_key, fields=FIELDS, lang='en'):
        Resolver.__init__(self, fields=fields, lang='en')
        self.uri = 'https://pro.ip-api.com/'
        self.api_key = api_key
        self.enable_sleep = False
        logger.info("API key found.")



"""Plain-text parsing tests"""
import unittest
import os

from libchickadee.parsers.plain_text import PlainTextParser

__author__ = 'Chapin Bryce'
__date__ = 20200107
__license__ = 'GPLv3 Copyright 2019 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class PlainTextParserTestCase(unittest.TestCase):
    """Plain-text parsing tests"""
    def setUp(self):
        """Test config"""
        self.test_data_ips = {
            '10.0.1.2': 1,
            '8.8.8.8': 1,
            '1.1.1.1': 2,
            '2.2.2.2': 1,
            '4.4.4.4': 1,
            '2001:4860:4860::8844': 1,
            '2001:4860:4860::8888': 1,
            '2001:4860:4860:0:0:0:0:8844': 1,
            '2001:4860:4860:0:0:0:0:8888': 1
        }
        self.parser = PlainTextParser()
        self.test_data_dir = os.path.join(
            os.path.dirname(__file__), 'test_data')

    def test_ip_extraction_txt(self):
        """Test text file extraction"""
        self.parser.parse_file(self.test_data_dir+'/txt_ips.txt')
        self.assertEqual(self.test_data_ips, self.parser.ips)

    def test_ip_extraction_gz(self):
        """Test GZ Text file extraction"""
        self.parser.parse_file(self.test_data_dir+'/txt_ips.txt.gz')
        self.assertEqual(self.test_data_ips, self.parser.ips)

    def test_gz_gzip_detection(self):
        """Test GZ detection"""
        self.assertTrue(
            self.parser.is_gz_file(self.test_data_dir+'/txt_ips.txt.gz')
        )

    def test_gz_txt_detection(self):
        """Test GZ detection"""
        self.assertFalse(
            self.parser.is_gz_file(self.test_data_dir+'/txt_ips.txt')
        )


if __name__ == '__main__':
    unittest.main()



"""Chickadee

An application to provide context for an IP address

"""

import argparse
import os
import sys
import logging
from pathlib import PurePath
from collections import Counter
import _io

# Third Party Libs
from tqdm import tqdm

# Import Backends
from libchickadee.backends.ipapi import Resolver, ProResolver

# Import Parsers
from libchickadee.parsers.plain_text import PlainTextParser
from libchickadee.parsers.xlsx import XLSXParser


__author__ = 'Chapin Bryce'
__date__ = 20200114
__license__ = 'GPLv3 Copyright 2019 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.

Will use the free rate limited ip-api.com service for resolution.
Please set an environment variable named CHICKADEE_API_KEY with the
value of your API key to enabled unlimited requests with the
commercial API
'''

logger = logging.getLogger(__name__)
_FIELDS = ','.join([  # Ordered list of fields to gather
    'query',
    'count', 'as', 'org', 'isp',
    'continent', 'country', 'regionName', 'city', 'district', 'zip',
    'mobile', 'proxy', 'hosting', 'reverse',
    'lat', 'lon', 'timezone',
    'status', 'message'
])


class Chickadee(object):
    """Class to handle chickadee script operations."""
    def __init__(self, outformat='json', outfile=sys.stdout, fields=_FIELDS):
        self.input_data = None
        self.outformat = outformat
        self.outfile = outfile
        self.fields = fields
        self.force_single = False
        self.lang = 'en'
        self.pbar = False
        self.resolve_ips = True

    def run(self, input_data):
        """Evaluate the input data format to extract and resolve IP addresses.

        Args:
            input_data (str or file_obj): User provided data containing IPs to
                resolve

        Returns:
            (list): List of dictionaries containing resolved hits
        """
        self.input_data = input_data
        results = []
        result_dict = {}
        # Extract and resolve IP addresses
        if not isinstance(self.input_data, _io.TextIOWrapper) and \
                os.path.isdir(self.input_data):
            logger.debug("Detected the data source as a directory")
            result_dict = self.dir_handler(self.input_data)  # Dir handler
        elif isinstance(self.input_data, _io.TextIOWrapper) or \
                os.path.isfile(self.input_data):
            logger.debug("Detected the data source as a file")
            result_dict = self.file_handler(self.input_data)  # File handler
        elif isinstance(self.input_data, str):
            logger.debug("Detected the data source as raw value(s)")
            result_dict = self.str_handler(self.input_data)  # String handler

        # Resolve if requested
        if self.resolve_ips:
            results = self.resolve(result_dict)
            return results

        return [{'query': k, 'count': v, 'message': 'No resolve'}
                for k, v in result_dict.items()]

    def write_output(self, results):
        """Write results to output format and/or files

        Args:
            results (list): List of GeoIP results
        """

        if self.outformat == 'csv':
            logger.debug("Writing CSV report")
            Resolver.write_csv(self.outfile, results, self.fields)
        elif self.outformat == 'json':
            logger.debug("Writing json report")
            Resolver.write_json(self.outfile, results, self.fields)
        elif self.outformat == 'jsonl':
            logger.debug("Writing json lines report")
            Resolver.write_json(self.outfile, results, self.fields, lines=True)

    @staticmethod
    def get_api_key():
        """Retrieve an API key set as an envar."""
        api_key = os.environ.get('CHICKADEE_API_KEY', None)
        if api_key is not None and len(api_key):
            return api_key
        return None

    @staticmethod
    def str_handler(data):
        """Handle string input of one or more IP addresses and executes the query

        Args:
            input_data (str): raw input data from user
            fields (list): List of fields to query for

        Return:
            data_dict (dict): dictionary of distinct IP addresses to resolve.
        """
        if isinstance(data, str) and ',' in data:
            # List of IPs
            raw_data = data.strip().split(',')
        elif isinstance(data, str):
            # Single IP
            raw_data = [data.strip()]
        else:
            raise TypeError('Unsupported input provided.')

        # Generate a distinct list with count
        data_dict = {}
        for x in raw_data:
            if x not in data_dict:
                data_dict[x] = 0
            data_dict[x] += 1
        return data_dict

    def resolve(self, data_dict):
        """Resolve IP addresses stored as keys within `data_dict`. The values
        for each key should represent the number of occurances of an IP within
        a data set.

        Args:
            data_dict (dict): Structured as {'IP': COUNT}

        Returns:
            all_results (list): resolved GeoIP data
        """
        distinct_ips = list(data_dict.keys())

        logger.info("Identified {} distinct IPs for resolution".format(
            len(distinct_ips)))

        api_key = self.get_api_key()

        if api_key:
            resolver = ProResolver(api_key, fields=self.fields, lang=self.lang)
        else:
            resolver = Resolver(fields=self.fields, lang=self.lang)

        if self.pbar:
            resolver.pbar = self.pbar

        logger.info("Resolving IPs")
        if self.force_single:
            results = []
            data = distinct_ips
            if self.pbar:
                data = tqdm(distinct_ips, desc="Resolving IPs",
                            unit_scale=True)

            for element in data:
                resolver.data = element
                results.append(resolver.single())
        else:
            results = resolver.query(distinct_ips)

        logger.info("Resolved IPs")

        # Add frequency information to results
        if 'count' in self.fields:
            updated_results = []
            for result in results:
                query = str(result.get('query', ''))
                result['count'] = int(data_dict.get(query, '0'))
                updated_results.append(result)

            return updated_results
        return results

    @staticmethod
    def file_handler(file_path):
        """Handle parsing IP addresses from a file

        Args:
            input_data (str): raw input data from use
            fields (list): List of fields to query for

        Return:
            (list): all query results from extracted IPs
        """
        if isinstance(file_path, _io.TextIOWrapper):
            is_stream = True
            # Extract IPs with proper handler
            logger.debug("Extracting IPs from STDIN")
        else:
            is_stream = False
            # Extract IPs with proper handler
            logger.debug("Extracting IPs from {}".format(file_path))

        if not is_stream and file_path.endswith('xlsx'):
            file_parser = XLSXParser()
        else:
            file_parser = PlainTextParser()
        try:
            file_parser.parse_file(file_path, is_stream)
        except Exception:
            logger.warning("Failed to parse {}".format(file_path))
        return file_parser.ips

    def dir_handler(self, file_path):
        """Handle parsing IP addresses from files recursively

        Args:
            input_data (str): raw input data from use
            fields (list): List of fields to query for

        Return:
            (list): all query results from extracted IPs
        """
        result_dict = {}
        for root, _, files in os.walk(file_path):
            for fentry in files:
                file_entry = os.path.join(root, fentry)
                logger.debug("Parsing file {}".format(file_entry))
                file_results = self.file_handler(file_entry)
                logger.debug("Parsed file {}, {} results".format(
                    file_entry, len(file_results)))
                result_dict = dict(Counter(result_dict)+Counter(file_results))
        logger.debug("{} total distinct IPs discovered".format(
            len(result_dict)))
        return result_dict


def setup_logging(path, verbose=False):
    """Function to setup logging configuration and test it."""
    # Allow us to modify the `logger` variable within a function
    global logger

    # Set logger object, uses module's name
    logger = logging.getLogger(name=__name__)

    # Set default logger level to DEBUG. You can change this later
    logger.setLevel(logging.DEBUG)

    # Logging formatter. Best to keep consistent for most usecases
    log_format = logging.Formatter(
        '%(asctime)s %(filename)s %(levelname)s %(module)s '
        '%(funcName)s:%(lineno)d - %(message)s')

    # Setup STDERR logging, allowing you uninterrupted
    # STDOUT redirection
    stderr_handle = logging.StreamHandler(stream=sys.stderr)
    if verbose:
        stderr_handle.setLevel(logging.DEBUG)
    else:
        stderr_handle.setLevel(logging.INFO)
    stderr_handle.setFormatter(log_format)

    # Setup file logging
    file_handle = logging.FileHandler(path, 'a')
    file_handle.setLevel(logging.DEBUG)
    file_handle.setFormatter(log_format)

    # Add handles
    logger.addHandler(stderr_handle)
    logger.addHandler(file_handle)


class CustomArgFormatter(argparse.RawTextHelpFormatter,
                         argparse.ArgumentDefaultsHelpFormatter):
    """Custom argparse formatter class"""


def arg_handling():
    """Argument handling."""
    parser = argparse.ArgumentParser(
        description=__desc__,
        formatter_class=CustomArgFormatter,
        epilog="Built by {}, v.{}".format(__author__, __date__)
    )
    parser.add_argument(
        'data',
        help="Either an IP address, comma delimited list of IP addresses, "
             "or path to a file or folder containing files to check for "
             "IP address values. Currently supported file types: "
             "plain text (ie logs, csv, json), gzipped plain text, xlsx "
             "(must be xlsx extension). Can accept plain text data as stdin.",
        nargs='*',
        default=sys.stdin
    )
    parser.add_argument('-f', '--fields',
                        help='Comma separated fields to query',
                        default=_FIELDS)
    parser.add_argument('-t', '--output-format',
                        help='Output format',
                        choices=['json', 'jsonl', 'csv'],
                        default='jsonl')
    parser.add_argument('-w', '--output-file',
                        help='Path to file to write output',
                        default=sys.stdout, metavar='FILENAME.JSON')
    parser.add_argument('-n', '--no-resolve', action='store_false',
                        help="Only extract IP addresses, don't resolve.")
    parser.add_argument('-s', '--single',
                        help="Use the significantly slower single item API. "
                             "Adds reverse DNS.",
                        action='store_true')
    parser.add_argument('--lang', help="Language", default='en',
                        choices=['en', 'de', 'es', 'pt-BR', 'fr', 'ja',
                                 'zh-CN', 'ru'])
    parser.add_argument('-p', '--progress', help="Enable progress bar",
                        action="store_true")
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Include debug log messages')
    parser.add_argument('-V', '--version', action='version',
                        help='Displays version',
                        version=str(__date__))
    parser.add_argument(
        '-l',
        '--log',
        help="Path to log file",
        default=os.path.abspath(os.path.join(
            os.getcwd(), PurePath(__file__).name.rsplit('.', 1)[0] + '.log'))
    )
    args = parser.parse_args()
    return args


def entry(args=None):
    """Entrypoint for package script"""
    args = arg_handling()
    fields = args.fields.split(',')
    setup_logging(args.log, args.verbose)
    logger.info("Starting Chickadee")
    for arg in vars(args):
        logger.debug("Argument {} is set to {}".format(
            arg, getattr(args, arg)
        ))
    logger.info("Configuring Chickadee")
    chickadee = Chickadee(fields=fields)
    chickadee.resolve_ips = args.no_resolve
    chickadee.force_single = args.single
    chickadee.lang = args.lang
    chickadee.pbar = args.progress

    logger.info("Parsing input")
    if isinstance(args.data, list):
        data = []
        for x in args.data:
            res = chickadee.run(x)
            data += res
    else:
        data = chickadee.run(args.data)

    logger.info("Writing output")
    chickadee.outfile = args.output_file
    chickadee.outformat = args.output_format
    chickadee.write_output(data)

    logger.info("Chickadee complete")


if __name__ == "__main__":
    entry()



"""Parse IP addresses from XLSX files
and feed to the Chickadee GeoIP API
"""

import os

from openpyxl import load_workbook

from libchickadee.parsers import IPv4Pattern, IPv6Pattern, strip_ipv6


__author__ = 'Chapin Bryce'
__date__ = 20200107
__license__ = 'GPLv3 Copyright 2019 Chapin Bryce'
__desc__ = '''Yet another GeoIP resolution tool.'''


class XLSXParser(object):
    """Class to extract IP addresses from xlsx workbooks."""
    def __init__(self):
        self.ips = dict()

    def parse_file(self, file_entry, is_stream=False):
        """Parse xlsx contents"""
        wb = load_workbook(file_entry)

        for sheet in wb.sheetnames:
            ws = wb[sheet]
            for row in ws.iter_rows():
                for cell in row:
                    if isinstance(cell.value, (str, bytes)):
                        self.check_ips(cell.value)

    def check_ips(self, data):
        """Check data for IP addresses."""
        for ipv4 in IPv4Pattern.findall(data):
            if ipv4 not in self.ips:
                self.ips[ipv4] = 0
            self.ips[ipv4] += 1
        for ipv6 in IPv6Pattern.findall(data):
            if strip_ipv6(ipv6) not in self.ips:
                self.ips[strip_ipv6(ipv6)] = 0
            self.ips[strip_ipv6(ipv6)] += 1


if __name__ == '__main__':  # pragma: no cover
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('path', help="File or folder to parse")
    args = parser.parse_args()

    xlparser = XLSXParser()
    if os.path.isdir(args.path):
        for root, _, files in os.walk(args.path):
            for fentry in files:
                xlparser.parse_file(os.path.join(root, fentry))
    else:
        xlparser.parse_file(args.path)
    print("{} unique IPs discovered".format(len(xlparser.ips)))
